##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool
$Id: LinkCheckerTool.py,v 1.4 2003/11/15 17:32:51 ctheune Exp $"""

# Python stuff
from types import StringTypes

# Global Zope stuff
from DateTime import DateTime
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass, PersistentMapping, Persistent
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.CMFCorePermissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

# Sibling imports
from interfaces import ILinkCheckerTool, ILinkInformation
from LinkInfo import LinkInfo
import retrievers

import zLOG

# Threading stuff
from threading import Thread
try:
    from threading import BoundedSemaphore
except ImportError:
    from threading import Semaphore
    BoundedSemaphore = Semaphore
import urllib, httplib
import httpclient
import Queue
import time

# Threading and transactions are a pain
import Globals
import OFS.Application
import ZODB.ZApplication

# Link checker stuff
from httplib import ResponseNotReady
from httpcodes import http_codes

retrieval_queue = Queue.Queue(0) # Make this an infinite synchronized queue
thread_stop = 0

active_threads = 5
thread_count = 0
pool_threads = BoundedSemaphore(value=active_threads)

class TooManyThreadsRunning(Exception):
    pass

class UnknownLink(Exception):
    pass

class LinkCheckerTool(UniqueObject, SimpleItem):
    """A link checker tool. Encapsulating mechanisms for finding links in the
       site, testing their availability and notifying their owners."""

    __implements__ = (ILinkCheckerTool, )

    id = 'portal_linkchecker'
    meta_type = 'CMF Linkchecker Tool'
    security = ClassSecurityInfo()
    plone_tool = 1

    manage_options = ( ({'label' : 'Overview', 'action' : 'manage_overview'},
                        {'label' : 'Configure', 'action' : 'manage_configure'},
                        {'label' : 'Linklist', 'action' : 'manage_checklinks'},
                        {'label' : 'Linkchecking', 'action' : 'manage_checker'},

                        ) + SimpleItem.manage_options )

    maximumCacheAge = 60*60     # 1 hour
    gracefulUpdateAge = 60*45   # 45 minutes
    defaultURIPrefix = ""       # No prefix by default

    #
    # ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = PageTemplateFile('www/explainLinkcheckerTool', globals(),
                                        __name__='manage_overview')

    security.declareProtected(ManagePortal, 'manage_configure')
    manage_configure= PageTemplateFile('www/configureLinkcheckerTool', globals(),
                                        __name__='manage_configure')

    security.declareProtected(ManagePortal, 'manage_checklinks')
    manage_checklinks = PageTemplateFile('www/checkLinks', globals(),
                                        __name__='manage_checklinks')
         
    security.declareProtected(ManagePortal, 'manage_checker')
    manage_checker = PageTemplateFile('www/checker', globals(),
                                        __name__='manage_checker')
                                        
    security.declareProtected(ManagePortal, 'edit_configuration')
    def edit_configuration(self, maxage, graceage, defaultURIPrefix):
        """Update the configuration options."""
        self.setMaximumCacheAge(maxage)
        self.setGracefulUpdateAge(graceage)
        self.setDefaultURIPrefix(defaultURIPrefix)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_configure')

    security.declareProtected(ManagePortal, 'manage_checkLink')
    def manage_checkLink(self, url):
        """ZMI method for checking a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
            return           
        data = self._cache[url]
        self.checkLink(data)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Link+checked')

    def manage_forceCheckLink(self, url):
        """ZMI method for forcing the check of a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
            return
        data = self._cache[url]
        self.__queueAddLink(data)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Check+forced')

    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_checkAll(self):
        """ZMI method for checking all links."""
        self.checkAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+checked+or+queued.')
        
    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_forceCheckAll(self):
        """ZMI method for checking all links."""
        self.forceCheckAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+forced+into+que.')

    security.declareProtected(ManagePortal, 'manage_findLinks')
    def manage_findLinks(self):
        """ZMI method for finding new links."""
        count = self.findLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checklinks?manage_tabs_message=%s+links+added.+%s+links+removed.' % count) 
            
    def manage_activateWorkerThread(self):
        """ZMI method for activating a new worker thread."""
        try:
            self.activateWorkerThread()
            message = "Activated worker thread."
        except TooManyThreadsRunning:
            message = "Already %s running." % active_threads
            
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checker?manage_tabs_message=%s' % message)

    #
    # Interface methods
    #

    def __init__(self):
        self._retrievers = PersistentMapping()
        self._cache = PersistentMapping()

        # register all global retrievers
        for type, retriever in retrievers.retrievers.items():
            self.registerRetriever(type, retriever)

    ## registry functions for link information retrieval

    security.declareProtected(ManagePortal, 'registerRetriever')
    def registerRetriever(self, portal_type, function_ref):
        assert not self._retrievers.has_key(portal_type), \
            "Duplicate portal_type registration (%s)" % portal_type
        self._retrievers[portal_type] = function_ref

    def getRetrievableTypes(self):
        return self._retrievers.keys()

    def getRetriever(self, type):
        return self._retrievers[type]
        
    security.declareProtected(ManagePortal, 'checkLink')
    def checkLink(self, data):
        if not self._cache.has_key(data.getCacheKey()):
            raise UnknownLink

        now = time.time()
        if data.lastcheck:
            age = data.lastcheck.millis() / 1000
        else:
            age = 0
        
        # Cache invalid? Requeue an update and return "unknown"
        if now - age > self.maximumCacheAge:
            self.__queueAddLink(data)
            return "unknown"

        # Update required?
        if now - age > self.gracefulUpdateAge:
            self.__queueAddLink(data)
        
        return data.state

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def checkAllLinks(self):
        for link in self._cache.values():
            self.checkLink(link)

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def forceCheckAllLinks(self):
        for link in self._cache.values():
            self.__queueAddLink(link)

    security.declareProtected(ManagePortal, 'findLinks')
    def findLinks(self):
        count = 0
        links = []
        stale = self._cache.keys()
        
        # gather all objects that are of a type we can check for links
        for key in self.getRetrievableTypes():
            objects = self.portal_catalog(portal_type=key)
            retriever = self.getRetriever(key)
            for ob in objects:
                ob = ob.getObject()
                for link in retriever(ob):
                    if not self._cache.has_key(link.getCacheKey()):
                        self._cache[link.getCacheKey()] = link
                        count = count + 1
                    if link.getCacheKey() in stale:
                        del stale[stale.index(link.getCacheKey())]
        # Update the URLS, they may need context. XXX this is icky
        for ob in self._cache.values():
            ob = ob.__of__(self)
            oldhash = ob.getCacheKey()
            ob.setUrl(ob.url, self.REQUEST)
            if oldhash != ob.getCacheKey():
                del self._cache[oldhash]
                self._cache[ob.getCacheKey()] = ob
            if ob.getCacheKey() in stale:
                del stale[stale.index(ob.getCacheKey())]
                
        # Everything in stale is stale. Kick it from the cache
        stalecount = len(stale)
        for x in stale:
            del self._cache[x]
        
        return count, stalecount

    security.declareProtected(ManagePortal, 'showLinks')
    def showLinks(self):
        """Show a grouped list of link data."""
        data = self._cache.values()

        # group them all
        grouped = {}
        for x in data:
            group = x.getGroupId()
            if not grouped.has_key(group):
                grouped[group] = []
            grouped[group].append(x)
        return grouped
        
        return self._cache.values()
                
    def checkWholeSite(self):
        for link in self.findLinks():
            self.checkLink(link)

    security.declareProtected(ManagePortal, 'setMaximumCacheAge')
    def setMaximumCacheAge(self, age):
        self.maximumCacheAge = age

    security.declareProtected(ManagePortal, 'setGracefulUpdateAge')
    def setGracefulUpdateAge(self, age):
        assert age < self.maximumCacheAge, \
            "Graceful update age is larger than maximum cache age."
        self.gracefulUpdateAge = age
     
    security.declareProtected(ManagePortal, 'setDefaultURIPrefix')
    def setDefaultURIPrefix(self, defaultURIPrefix):
        assert type(defaultURIPrefix) in StringTypes
        self.defaultURIPrefix = defaultURIPrefix
        
    security.declareProtected(ManagePortal, 'queueSize')
    def queueSize(self):
        global retrieval_queue
        return retrieval_queue.qsize()

    security.declareProtected(ManagePortal, 'activateWorkerThread')
    def activateWorkerThread(self):
        if not pool_threads.acquire(0):
            raise TooManyThreadsRunning
        global thread_count
        thread_count += 1
        thread = URLCheckerThread(retrieval_queue)
        thread.start()

    security.declareProtected(ManagePortal, 'countWorkingThreads')
    def countWorkingThreads(self):
        global thread_count
        return thread_count
        
    ### Private methods
    security.declarePrivate('__queueAddLink')
    def __queueAddLink(self, info):
        """Schedules a new check.

        Info must be a ILinkInfo object."""
        retrieval_queue.put((self.getPhysicalPath(), info.getCacheKey())) 
        info.queued = 1

InitializeClass(LinkCheckerTool)

#### Threading code.
# We only use one worker thread. Object references
# should be stable to work with.

class URLCheckerThread(Thread):

    def __init__(self, queue):
        Thread.__init__(self, name="URLChecker")
        self.queue = queue

    def run(self):
        zLOG.LOG('CMFLinkChecker', 0, 'Initialized', '\tWorker thread has been initialized.\n')
        # try to get a zodb connection
        check_app = None
        while 1:
            if check_app:
                get_transaction().commit()
                check_app._p_jar.sync()
                check_app._p_jar.close()
                check_app = None

            # we are doing cooperative multitasking
            # this thread doesn't need too much power
            time.sleep(0.1)
            try:
                linkchecker, key = self.queue.get(block=0)
            except Queue.Empty:
                time.sleep(1)
                continue
            
            # We got work, so setup the database connection.
            DB=Globals.DB
            app=ZODB.ZApplication.ZApplicationWrapper(
                DB, 'Application', OFS.Application.Application, (),
                Globals.VersionNameName)
            check_app=app()
            
            o = check_app
            # find the linkchecker object
            for p in linkchecker:
                if p:
                    o=getattr(o,p)
            linkchecker = o

            item = linkchecker._cache[key]

            # Links need to revalidate everytime otherwise they are in alert state.
            item.state_class = "alert"
            item.queued = 0
            item.lastcheck = DateTime()

            # only check http links. other links are ... hmm
            # ... unknown ...
            type, url = urllib.splittype(item.url)
            if not type == 'http':
                item.state = 'unknown'
                item.state_class = 'unknown'
                item.lasttime = 0
                item.headers = {}
                continue
            
            # get a connection to the server
            host, path = urllib.splithost(url)
            if host is None:
                item.state = "invalid"
                item.state_class = "unknown"
                item.reason = "Invalid Link: No hostname defined"
                continue
            host, port = urllib.splitnport(host, 80)
            
            conn = httpclient.myHTTPConnection(host, port, 2)
            starttime = time.time()
            # request the link
            try:
                conn.request('GET', path)
            except:
                import sys
                item.state = "unavailable"
                item.reason = sys.exc_value
                continue

            # look into the response
            response = None
            while not response:
                try:
                    now = time.time()
                    response = conn.getresponse()
                except ResponseNotReady:
                    if starttime - now > 4:
                        break
                time.sleep(0.2)

            conn.close()
            response.close()

            item.lasttime = now - starttime
            item.headers = {}

            # what if we didn't got a response?
            if not response:
                item.state = 'failed'
                continue
            item.state = response.status

            if not http_codes.has_key(item.state):
                item.state_class = 'unknown'
                continue

            translated_code = http_codes.get(item.state)
            if translated_code[0]:
                item.state_class="green"
            item.reason = translated_code[1]
                
        zLOG.LOG('CMFLinkChecker', 0, 'Stopped', '\tWorker thread stopped.\n')
        global thread_count
        thread_count -= 1

